#!/usr/bin/env ruby
require 'getopts'
require 'fileutils'


class JpegParser
  attr_accessor :error
  APP7 = 0xe7

  def initialize(filename)
    @segments = []
    @data = nil
    @error = nil
    @zco = nil

    open(filename) do |f|
      @data = f.read
    end

    while @data.length > 0 && !@error 
      get_segment();
    end
  end

  def zco()
    unless @zco
      content = content_of(APP7)
      if content
        @zco = content[4..-1]
      end
    end
    @zco
  end

  def width()
    content = content_of(0xc0)
    content ? content[7..9].unpack("n").first : 0
  end

  def height()
    content = content_of(0xc0)
    content ? content[5..7].unpack("n").first : 0
  end

  def content_of(marker)
    @segments.each do |segment|
      return segment[:content] if segment[:marker] == marker
    end
    nil
  end

  def get_segment()
    if @data[0] == 0xff
      marker = @data[1]
      case marker
      when 0xd8, 0xd9
        size = 0
        content = @data.slice!(0, 2)

      when 0xda
        size = @data[2..3].unpack("n")[0]
        content = @data.slice!(0, size + 2)
        index = @data.index(/\377[^\0]/)
        if index
          content += @data.slice!(0, index)
          size += index
        else
          print "bad SOS segment.\n"
          @error = true
        end

      else
        size = @data[2..3].unpack("n")[0]
        content = @data.slice!(0, size + 2)
      end

      @segments.push :marker => marker, :size => size, :content => content
      #print "segment #{marker.to_s(16)} size #{size} \n"
      return

    else
      print "unknown segment found.\n"
      @error = true
    end
  end
end

################################################################

class Mqo
  attr_accessor :error

  def initialize(zco, params)
    @zco = zco
    @params = params
    @error = false

    @header = zco[0..28]
    @body = zco[28..-1]
    @version = @header[3]
    @segment_width, @segment_height = @header[6..10].unpack("vv")
    @data = @body.unpack("f*")

    @name = params[:name].gsub(/[^a-zA-Z0-9]/, "_") || "zco"
    @texture_name = params[:texture_name]
    @width = params[:width]
    @height = params[:height]

    @num_vertices = (@segment_width + 1) * (@segment_height + 1)
    @num_faces = @segment_width * @segment_height * 2

    @error = true unless @header[0..2] == "ZCO"
    @error = true unless @num_vertices == @data.size
  end

  def dump
    result = []
    result.push "Metasequoia Document"
    result.push "Format Text Ver 1.0"
    result.push ""

    dump_material(result)
    dump_object(result)

    result.push ""
    result.join "\r\n"
  end

  def dump_material(result)
    mat = "  zco shader(1) tex(#{@texture_name}) proj_type(0)"
    result.push "Material 1 {"
    result.push mat
    result.push "}"
    result.push ""
  end

  def dump_object(result)
    result.push "Object #{@name} {"
    dump_vertices(result)
    dump_faces(result)
    result.push "}"
    result.push ""
  end

  def dump_vertices(result)
    width = mesh_width
    height = mesh_height
    sw = @segment_width.to_f
    sh = @segment_height.to_f
    stride = @segment_width + 1
    scale = 100.0

    result.push "  vertex #{@num_vertices} {"

    (0...@segment_height + 1).each do |y|
      (0...@segment_width + 1).each do |x|
        result.push "    %.4f %.4f %.4f" % 
          [ width * (x / sw - 0.5) * scale,
            height * (y / sh - 0.5) * scale,
            @data[x + y * stride] * scale]
      end
    end
    result.push "  }"
  end

  def dump_faces(result)
    width = mesh_width
    height = mesh_height
    sw = @segment_width.to_f
    sh = @segment_height.to_f
    stride = @segment_width + 1
    result.push "  face #{@num_faces} {"

    (0...@segment_height).each do |y|
      (0...@segment_width).each do |x|
        i = x + y * (@segment_width + 1)
        u0 = x / sw
        u1 = (x + 1) / sw
        v0 = 1.0 - y / sh
        v1 = 1.0 - (y + 1) / sh
        format = "    3 V(%s) M(0) UV(%s) COL(0xffffffff)"

        vertices = "%d %d %d" % [ i, i + stride, i + 1 ]
        uvs = "%.5f %.5f %.5f %.5f %.5f %.5f" % [ u0, v0, u0, v1, u1, v0 ]
        result.push format % [ vertices, uvs ]

        vertices = "%d %d %d" % [ i + stride, i + stride + 1, i + 1]
        uvs = "%.5f %.5f %.5f %.5f %.5f %.5f" % [ u0, v1, u1, v1, u1, v0 ]
        result.push format % [ vertices, uvs ]
      end
    end
    result.push "  }"
  end

  def mesh_width
    (@width < @height) ? @width.to_f / @height : 1.0
  end

  def mesh_height
    (@width > @height) ? @height.to_f / @width : 1.0
  end

end

################################################################

def to_mqo(argv)
  name = argv.gsub(/(\.z)?\.jpg$/i, "")

  parser = JpegParser.new(argv)
  unless parser.error
    if parser.zco
      mqo = Mqo.new(parser.zco, 
                    :name => name,
                    :texture_name => argv,
                    :width => parser.width, 
                    :height => parser.height)

      unless mqo.error
        open("#{name}.mqo", "w") do |f|
          f.write(mqo.dump)
        end
      else
        print "invalid depth data.\n"
      end
    else
      print "no depth data included.\n"
    end
  else
    print "parse error.\n"
  end
end


if getopts('v', 'version')
  if $OPT_v or $OPT_version
    puts "1.0.0"
    exit
  end

  if ARGV.length > 0
    ARGV.each do |argv|
      to_mqo(argv)
    end
    exit
  end
end

abort "usage: #$0 [-v] [--version] filename.jpg\n"
